#####################################################################
### Proprify substitutions                                        ###
#####################################################################

# Some examples
chacon = WordMorphism("a->aaba,b->b")
tribo = WordMorphism({1:[1,2], 2:[1,3], 3:[1]})
sigma6 = WordMorphism({1:[1,2], 2:[3], 3:[4,3], 4:[5], 5:[6], 6:[1]})
hokkaido = WordMorphism({1:[1,2], 2:[3], 3:[4], 4:[5], 5:[1]})
bad1 = WordMorphism({1:[2,1,4], 2:[3], 3:[4], 4:[1]})
bad2 = WordMorphism({1:[4,2], 2:[4,3], 3:[4], 4:[1]})
bad3 = WordMorphism({1:[2,1,3], 2:[4], 3:[5], 4:[1], 5:[2,1]})
bad4 = WordMorphism({1:[3,4], 2:[5,3,4], 3:[5,3], 4:[4,6], 5:[1], 6:[2]})

""" Examples of usage:
sage: load proprifier.sage
sage: chacon = WordMorphism("a->aaba,b->b")
sage: print proprify(chacon)
    Return word used: a.a with sigma^1
    WordMorphism:
    1->846583127931279465831279465,
    2->8465831279465,
    3->8465831279465,
    4->8465831279465,
    5->8465831279465846583127931279465831279312...,
    6->846583127931279465831279465,
    7->8465831279312794658312793127946583127946...,
    8->8465831279465846583127931279465831279465...,
    9->8465831279312794658312793127946583127946...
sage: print proprify(chacon, ind=True)
    Return word used: a.a with sigma^1
    WordMorphism:
    1->12132,
    2->121332132,
    3->1213321332132
"""

def is_proper(sigma):
    imgs = [sigma(a) for a in sigma.domain().alphabet()]
    L = set(w[0] for w in imgs)
    R = set(w[-1] for w in imgs)

    if len(L) == 1 or len(R) == 1:
        return True
    else:
         return False

def proprify(sigma, letter=1, ind=False, fp_size=2000):
    r"""
    Input:
        - sigma

        - ind (default=False): if True return only the "induced" version of sigma (tau)

        - fp_size (default=2000): size of the prefix of
            the fixed point in which to look for return words
            (shoud be replaced by an actual computation of the return words)
            Some very nasty substitutions need more that 2000,
            for example 1->23, 2->33, 3->4, 4->1 needs about 20000...

    Output: a proper substitution zeta such that X_sigma and X_zeta are isomorphic

                          sigma
            --------> A ---------> A <--------
            |         ^            ^         |
            |     phi |            | phi     |
            |         |   tau      |         |
      theta |         R ---------> R         | theta
            |         |            |         |
            |   gamma |            | gamma   |
            |         v   zeta     v         |
            |-------- B ---------> B --------|

    Reference:
        Chapter 6: Section 6.5.2.3, Proposition 6.5.4, Section 6.5.2.4
        Combinatorics, Automata and Number Theory, Chapter 6,
        Series Encyclopedia of Mathematics and its applications 135,
        Cambridge University Press 2010, 338-386
    """

    # look for l and r such that sigma(l) = l... and sigma(r) = ...r
    # and such that the word rl is contained in a fixed point
    # we might have to take a power of sigma
    power = 1
    found_rl = False

    while not found_rl:
        try:
            L = (sigma^power).fixed_points()
            R = (sigma^power).reversal().fixed_points()
            for x in L:
                for y in R:
                    if x.length() is Infinity and y.length() is Infinity:
                        l = x[0]
                        r = y[0]
                        for z in L:
                            if Word([r,l]).is_subword_of(z[:fp_size]):
                                found_rl = True
                                break
                        if found_rl:
                            break

            if not found_rl:
                power += 1

        except IndexError:
            power += 1

    print "Return word used: %s.%s with sigma^%s"%(r,l,power)
    sigma = sigma^power

    # compute the return words
    w = sigma.fixed_point(l)[:fp_size]
    return_words = sorted(w.return_words(Word([r,l])))
    for i, w in enumerate(return_words):
        return_words[i] = return_words[i][1:] + Word([return_words[i][0]])

    # build phi : R = [1,2, ...] -> A = sigma.alphabet()^*
    phi = {}
    phi_rev = {}
    for i, w in enumerate(return_words):
        phi[i+1] = w
        phi_rev[w] = i+1

    phi = WordMorphism(phi)
    R = phi.domain().alphabet()

    # build tau : R -> R
    tau = {}
    for a in R:
        img = sigma(phi(a))
        img_coded = []
        while img:
            k = img.rfind(Word([r,l])) + 1
            img_coded = [phi_rev[img[k:]]] + img_coded
            img = img[:k]
        tau[a] = img_coded

    tau = WordMorphism(tau)

    # make sure that the images of tau are larger than the images of phi
    power = 1
    power_enough = False
    while not power_enough:
        power_enough = True

        for a in R:
            if len((tau^power)(a)) < len(phi(a)):
                power += 1
                power_enough = False

    #print "2nd power: %d"%power
    tau = tau^power

    # return tau if we just want the induction substitution
    if ind:
        return tau

    # build gamma : R -> B
    gamma = {}
    for a in R:
        gamma[a] = Word([(a,i+1) for i in range(len(phi(a)))])

    gamma = WordMorphism(gamma)

    # build zeta : B -> B
    zeta = {}
    for a in R:
        img = tau(a)
        for i in range(len(phi(a))):
            if i + 1 < len(phi(a)):
                zeta[(a,i+1)] = gamma(img[i])
            else:
                zeta[(a,i+1)] = gamma(img[i:])

    # translate zeta with a nicer alpabet
    t = {}
    for i, x in enumerate(zeta.keys()):
        t[x] = i+1
    t = WordMorphism(t)

    zeta_translated = WordMorphism({i+1:t(zeta[x]) for i, x in enumerate(zeta.keys())})

    # make sure the final result is proper
    power = 1
    while not is_proper(zeta_translated^power):
        power += 1

    return zeta_translated^power
